@@ -45,6 +45,104 @@
 #include "../gamelink/gamelink.h"
 #endif // C_GAMELINK
 
+// ACPI memory region allocation.
+// Most ACPI BIOSes actually use some region at top of memory, but the
+// design of DOSBox-X doesn't make that possible, so the ACPI tables are
+// written to a high memory region just below the top 4GB region and the
+// RSD PTR in the legacy BIOS region (0xE0000-0xFFFFF) will point at that.
+// A memory address is chosen, which must be maintained once tables are
+// generated because tables point at each other by physical memory address.
+// A fixed size region is chosen within which the tables are written.
+//
+// NTS: ACPI didn't happen until the Pentium era when it became quite rare
+//      for CPUs to have less than 32 address bits. No 26-bit 486SX limits
+//      here. For this reason, ACPI is not supported unless all 32 address
+//      bits are enabled.
+bool ACPI_enabled = false;
+bool acpi_mem_setup = false;
+uint32_t ACPI_BASE=0;
+uint32_t ACPI_REGION_SIZE=0; // power of 2
+uint32_t ACPI_version=0;
+unsigned char *ACPI_buffer=NULL;
+size_t ACPI_buffer_size=0;
+int ACPI_IRQ=-1;
+unsigned int ACPI_SMI_CMD=0;
+
+class ACPIPageHandler : public PageHandler {
+	public:
+		ACPIPageHandler() : PageHandler(PFLAG_NOCODE|PFLAG_READABLE|PFLAG_WRITEABLE) {}
+		ACPIPageHandler(Bitu flags) : PageHandler(flags) {}
+		HostPt GetHostReadPt(Bitu phys_page) override {
+			assert(ACPI_buffer != NULL);
+			assert(ACPI_buffer_size >= 4096);
+			phys_page -= (ACPI_BASE >> 12);
+			phys_page &= (ACPI_REGION_SIZE >> 12) - 1;
+			if (phys_page >= (ACPI_buffer_size >> 12)) phys_page = (ACPI_buffer_size >> 12) - 1;
+			return ACPI_buffer + (phys_page << 12);
+		}
+		HostPt GetHostWritePt(Bitu phys_page) override {
+			assert(ACPI_buffer != NULL);
+			assert(ACPI_buffer_size >= 4096);
+			phys_page -= (ACPI_BASE >> 12);
+			phys_page &= (ACPI_REGION_SIZE >> 12) - 1;
+			if (phys_page >= (ACPI_buffer_size >> 12)) phys_page = (ACPI_buffer_size >> 12) - 1;
+			return ACPI_buffer + (phys_page << 12);
+		}
+};
+
+static ACPIPageHandler acpi_mem_handler;
+
+PageHandler* acpi_memio_cb(MEM_CalloutObject &co,Bitu phys_page) {
+	(void)co;//UNUSED
+	(void)phys_page;//UNUSED
+
+	if (ACPI_buffer != NULL && ACPI_REGION_SIZE != 0 && phys_page >= (ACPI_BASE/4096) && phys_page < ((ACPI_BASE+ACPI_REGION_SIZE)/4096))
+		return &acpi_mem_handler;
+
+	return NULL;
+}
+
+void MEM_ResetPageHandler_Unmapped(Bitu phys_page, Bitu pages);
+
+void ACPI_mem_enable(const bool enable) {
+	if (enable && !acpi_mem_setup) {
+		if (ACPI_BASE != 0 && ACPI_REGION_SIZE != 0) {
+			MEM_SetPageHandler( ACPI_BASE/4096, ACPI_REGION_SIZE/4096, &acpi_mem_handler );
+			acpi_mem_setup = true;
+			PAGING_ClearTLB();
+		}
+	}
+	else if (!enable && acpi_mem_setup) {
+		if (ACPI_BASE != 0 && ACPI_REGION_SIZE != 0) {
+			MEM_ResetPageHandler_Unmapped( ACPI_BASE/4096, ACPI_REGION_SIZE/4096 );
+			acpi_mem_setup = false;
+			PAGING_ClearTLB();
+		}
+	}
+}
+
+void ACPI_free() {
+	if (ACPI_buffer != NULL) {
+		delete[] ACPI_buffer;
+		ACPI_buffer = NULL;
+	}
+	ACPI_buffer_size = 0;
+}
+
+bool ACPI_init() {
+	if (ACPI_buffer == NULL) {
+		if (ACPI_REGION_SIZE == 0 || ACPI_REGION_SIZE > (8ul << 20ull))
+			return false;
+
+		ACPI_buffer_size = ACPI_REGION_SIZE;
+		ACPI_buffer = new unsigned char [ACPI_buffer_size];
+		if (ACPI_buffer == NULL)
+			return false;
+	}
+
+	return (ACPI_buffer != NULL);
+}
+
 static MEM_Callout_t lfb_mem_cb = MEM_Callout_t_none;
 static MEM_Callout_t lfb_mmio_cb = MEM_Callout_t_none;
 
@@ -113,11 +211,11 @@ HostPt MemBase = NULL;
 class UnmappedPageHandler : public PageHandler {
 public:
     UnmappedPageHandler() : PageHandler(PFLAG_INIT|PFLAG_NOCODE) {}
-    uint8_t readb(PhysPt addr) {
+    uint8_t readb(PhysPt addr) override {
         (void)addr;//UNUSED
         return 0xFF; /* Real hardware returns 0xFF not 0x00 */
     } 
-    void writeb(PhysPt addr,uint8_t val) {
+    void writeb(PhysPt addr,uint8_t val) override {
         (void)addr;//UNUSED
         (void)val;//UNUSED
     }
@@ -126,7 +224,7 @@ public:
 class IllegalPageHandler : public PageHandler {
 public:
     IllegalPageHandler() : PageHandler(PFLAG_INIT|PFLAG_NOCODE) {}
-    uint8_t readb(PhysPt addr) {
+    uint8_t readb(PhysPt addr) override {
         (void)addr;
 #if C_DEBUG
         LOG_MSG("Warning: Illegal read from %x, CS:IP %8x:%8x",addr,SegValue(cs),reg_eip);
@@ -139,7 +237,7 @@ public:
 #endif
         return 0xFF; /* Real hardware returns 0xFF not 0x00 */
     } 
-    void writeb(PhysPt addr,uint8_t val) {
+    void writeb(PhysPt addr,uint8_t val) override {
         (void)addr;//UNUSED
         (void)val;//UNUSED
 #if C_DEBUG
@@ -158,13 +256,13 @@ class RAMPageHandler : public PageHandler {
 public:
     RAMPageHandler() : PageHandler(PFLAG_READABLE|PFLAG_WRITEABLE) {}
     RAMPageHandler(Bitu flags) : PageHandler(flags) {}
-    HostPt GetHostReadPt(Bitu phys_page) {
+    HostPt GetHostReadPt(Bitu phys_page) override {
         if (!a20_fast_changeable || (phys_page & (~0xFul/*64KB*/)) == 0x100ul/*@1MB*/)
             return MemBase+(phys_page&memory.mem_alias_pagemask_active)*MEM_PAGESIZE;
 
         return MemBase+phys_page*MEM_PAGESIZE;
     }
-    HostPt GetHostWritePt(Bitu phys_page) {
+    HostPt GetHostWritePt(Bitu phys_page) override {
         if (!a20_fast_changeable || (phys_page & (~0xFul/*64KB*/)) == 0x100ul/*@1MB*/)
             return MemBase+(phys_page&memory.mem_alias_pagemask_active)*MEM_PAGESIZE;
 
@@ -177,10 +275,10 @@ public:
     ROMAliasPageHandler() {
         flags=PFLAG_READABLE|PFLAG_HASROM;
     }
-    HostPt GetHostReadPt(Bitu phys_page) {
+    HostPt GetHostReadPt(Bitu phys_page) override {
         return MemBase+((phys_page&0xF)+0xF0)*MEM_PAGESIZE;
     }
-    HostPt GetHostWritePt(Bitu phys_page) {
+    HostPt GetHostWritePt(Bitu phys_page) override {
         return MemBase+((phys_page&0xF)+0xF0)*MEM_PAGESIZE;
     }
 };
@@ -190,19 +288,19 @@ public:
     ROMPageHandler() {
         flags=PFLAG_READABLE|PFLAG_HASROM;
     }
-    void writeb(PhysPt addr,uint8_t val){
+    void writeb(PhysPt addr,uint8_t val) override {
         if (IS_PC98_ARCH && (addr & ~0x7FFF) == 0xE0000u)
             { /* Many PC-98 games and programs will zero 0xE0000-0xE7FFF whether or not the 4th bitplane is mapped */ }
         else
             LOG(LOG_CPU,LOG_ERROR)("Write %x to rom at %x",(int)val,(int)addr);
     }
-    void writew(PhysPt addr,uint16_t val){
+    void writew(PhysPt addr,uint16_t val) override {
         if (IS_PC98_ARCH && (addr & ~0x7FFF) == 0xE0000u)
             { /* Many PC-98 games and programs will zero 0xE0000-0xE7FFF whether or not the 4th bitplane is mapped */ }
         else
             LOG(LOG_CPU,LOG_ERROR)("Write %x to rom at %x",(int)val,(int)addr);
     }
-    void writed(PhysPt addr,uint32_t val){
+    void writed(PhysPt addr,uint32_t val) override {
         if (IS_PC98_ARCH && (addr & ~0x7FFF) == 0xE0000u)
             { /* Many PC-98 games and programs will zero 0xE0000-0xE7FFF whether or not the 4th bitplane is mapped */ }
         else
@@ -1621,7 +1719,7 @@ HostPt GetMemBase(void) { return MemBase; }
 class REDOS : public Program {
 public:
     /*! \brief      Program entry point, when the command is run */
-    void Run(void) {
+    void Run(void) override {
 		if (cmd->FindExist("/?", false) || cmd->FindExist("-?", false)) {
 			WriteOut("Reboots the kernel of DOSBox-X's emulated DOS.\n\nRE-DOS\n");
 			return;
@@ -1641,7 +1739,7 @@ void REDOS_ProgramStart(Program * * make) {
 class A20GATE : public Program {
 public:
     /*! \brief      Program entry point, when the command is run */
-    void Run(void) {
+    void Run(void) override {
         if (cmd->FindExist("-?", false) || cmd->FindExist("/?", false)) {
             WriteOut("Turns on/off or changes the A20 gate mode.\n\n");
             WriteOut("A20GATE [ON | OFF | SET [off | off_fake | on | on_fake | mask | fast]]\n\n"
@@ -1770,6 +1868,7 @@ void ShutDownRAM(Section * sec) {
 #endif
         MemBase = NULL;
     }
+    ACPI_free();
 }
 
 void MEM_InitCallouts(void) {
@@ -1826,7 +1925,7 @@ void Init_RAM() {
         Bitu maxsz;
 
         if (sizeof(void*) > 4) // 64-bit address space
-            maxsz = (Bitu)(3584ul * 1024ul); // 3.5GB
+            maxsz = (Bitu)(3968ul * 1024ul); // 3.9GB (up to 0xF8000000)
         else
             maxsz = (Bitu)(1024ul * 1024ul); // 1.0GB
 
@@ -2176,7 +2275,7 @@ public:
 	{}
 
 private:
-	virtual void getBytes(std::ostream& stream)
+	void getBytes(std::ostream& stream) override
 	{
 		uint8_t pagehandler_idx[0x40000];
 		unsigned int size_table;
@@ -2222,7 +2321,7 @@ private:
 		WRITE_POD( &pagehandler_idx, pagehandler_idx );
 	}
 
-	virtual void setBytes(std::istream& stream)
+	void setBytes(std::istream& stream) override
 	{
 		uint8_t pagehandler_idx[0x40000];
 		void *old_ptrs[4];
