@@ -55,7 +55,7 @@ extern bool shell_keyboard_flush;
 extern bool dos_shell_running_program, mountwarning, winautorun;
 extern bool startcmd, startwait, startquiet, internal_program;
 extern bool addovl, addipx, addne2k, enableime, showdbcs;
-extern bool halfwidthkana, force_conversion, gbk;
+extern bool halfwidthkana, force_conversion, gbk, uselangcp;
 extern const char* RunningProgram;
 extern int enablelfn, msgcodepage, lastmsgcp;
 extern uint16_t countryNo;
@@ -77,15 +77,15 @@ void initcodepagefont(void);
 void runMount(const char *str);
 void ResolvePath(std::string& in);
 void DOS_SetCountry(uint16_t countryNo);
-void SwitchLanguage(int oldcp, int newcp, bool confirm);
+bool SwitchLanguage(int oldcp, int newcp, bool confirm);
 void CALLBACK_DeAllocate(Bitu in), DOSBox_ConsolePauseWait();
 void GFX_SetTitle(int32_t cycles, int frameskip, Bits timing, bool paused);
-bool isDBCSCP(), InitCodePage(), isKanji1(uint8_t chr), shiftjis_lead_byte(int c), sdl_wait_on_error();
+bool isDBCSCP(), InitCodePage(), isKanji1(uint8_t chr), shiftjis_lead_byte(int c), sdl_wait_on_error(), CheckDBCSCP(int32_t codepage), TTF_using(void);
 
 Bitu call_shellstop = 0;
 /* Larger scope so shell_del autoexec can use it to
  * remove things from the environment */
-DOS_Shell * first_shell = 0; 
+DOS_Shell * first_shell = nullptr;
 
 static Bitu shellstop_handler(void) {
 	return CBRET_STOP;
@@ -275,7 +275,7 @@ DOS_Shell::DOS_Shell():Program(){
 	echo=true;
 	exit=false;
 	perm = false;
-	bf=0;
+	bf = nullptr;
 	call=false;
 	exec=false;
 	lfnfor = uselfn;
@@ -382,7 +382,7 @@ public:
     std::string str = "";
     unsigned long long curpos = 0;
 	device_TMP(char *name) { SetName(name); };
-	virtual bool Read(uint8_t * data,uint16_t * size) {
+	bool Read(uint8_t * data,uint16_t * size) override {
         int i;
         for (i=0; i<*size; i++) {
             if (curpos+i>=str.size()) break;
@@ -391,11 +391,11 @@ public:
 		*size = i;
 		return true;
 	}
-	virtual bool Write(const uint8_t * data,uint16_t * size) {
+	bool Write(const uint8_t * data,uint16_t * size) override {
         for (int i=0; i<*size; i++) str += std::string(1, data[i]);
 		return true;
 	}
-	virtual bool Seek(uint32_t * pos,uint32_t type) {
+	bool Seek(uint32_t * pos,uint32_t type) override {
 		switch (type) {
             case 0:
                 curpos = *pos;
@@ -411,13 +411,12 @@ public:
                 return false;
 		}
 		if (curpos > str.size()) curpos = str.size();
-		else if (curpos < 0) curpos = 0;
 		return true;
 	}
-	virtual bool Close() { return true; }
-	virtual uint16_t GetInformation(void) { return (strcmp(RunningProgram, "WCLIP") ? DeviceInfoFlags::Device : 0) | DeviceInfoFlags::EofOnInput; }
-	virtual bool ReadFromControlChannel(PhysPt bufptr,uint16_t size,uint16_t * retcode) { (void)bufptr; (void)size; (void)retcode; return false; }
-	virtual bool WriteToControlChannel(PhysPt bufptr,uint16_t size,uint16_t * retcode) { (void)bufptr; (void)size; (void)retcode; return false; }
+	bool Close() override { return true; }
+	uint16_t GetInformation(void) override { return (strcmp(RunningProgram, "WCLIP") ? DeviceInfoFlags::Device : 0) | DeviceInfoFlags::EofOnInput; }
+	bool ReadFromControlChannel(PhysPt bufptr,uint16_t size,uint16_t * retcode) override { (void)bufptr; (void)size; (void)retcode; return false; }
+	bool WriteToControlChannel(PhysPt bufptr,uint16_t size,uint16_t * retcode) override { (void)bufptr; (void)size; (void)retcode; return false; }
 };
 
 void DOS_Shell::ParseLine(char * line) {
@@ -427,10 +426,10 @@ void DOS_Shell::ParseLine(char * line) {
 	line = trim(line);
 
 	/* Do redirection and pipe checks */
-	
-	char * in  = 0;
-	char * out = 0;
-	char * toc = 0;
+
+	char * in  = nullptr;
+	char * out = nullptr;
+	char * toc = nullptr;
 
 	uint16_t dummy,dummy2;
 	uint32_t bigdummy = 0;
@@ -451,7 +450,7 @@ void DOS_Shell::ParseLine(char * line) {
 			DOS_OpenFile(in,OPEN_READ,&dummy);	//Open new stdin
 		} else {
 			WriteOut(!*in?"File open error\n":(dos.errorcode==DOSERR_ACCESS_DENIED?MSG_Get("SHELL_CMD_FILE_ACCESS_DENIED"):"File open error - %s\n"), in);
-			in = 0;
+			in = nullptr;
 			return;
 		}
 	}
@@ -686,7 +685,7 @@ const char *ParseMsg(const char *msg) {
 #if defined(USE_TTF)
         && halfwidthkana
 #endif
-        && InitCodePage() && dos.loaded_codepage==932) uselowbox = true;
+        && dos.loaded_codepage==932) uselowbox = true;
         force_conversion = false;
         dos.loaded_codepage=cp;
         if (uselowbox || IS_JEGA_ARCH || IS_JDOSV) {
@@ -817,6 +816,7 @@ void showWelcome(Program *shell) {
     }
 }
 
+bool finish_prepare = false;
 void DOS_Shell::Prepare(void) {
     if (this == first_shell) {
         Section_prop *section = static_cast<Section_prop *>(control->GetSection("dosbox"));
@@ -863,42 +863,29 @@ void DOS_Shell::Prepare(void) {
 				if (r!=NULL) *r=0;
 				country = atoi(trim(countrystr));
 				int32_t newCP = r==NULL||IS_PC98_ARCH||IS_JEGA_ARCH||IS_DOSV?dos.loaded_codepage:atoi(trim(r+1));
-                if (control->opt_langcp && msgcodepage>0 && isSupportedCP(msgcodepage) && msgcodepage != newCP)
-                    newCP = msgcodepage;
 				if (r!=NULL) *r=',';
                 if (!IS_PC98_ARCH&&!IS_JEGA_ARCH) {
-#if defined(USE_TTF)
-                    if (ttf.inUse) {
-                        if (newCP) {
-                            int missing = toSetCodePage(this, newCP, control->opt_fastlaunch?1:0);
-                            WriteOut(MSG_Get("SHELL_CMD_CHCP_ACTIVE"), dos.loaded_codepage);
-                            if (missing > 0) WriteOut(MSG_Get("SHELL_CMD_CHCP_MISSING"), missing);
-                        }
-                        else if (r!=NULL) WriteOut(MSG_Get("SHELL_CMD_CHCP_INVALID"), trim(r+1));
-                    } else
-#endif
-                    if (!newCP && IS_DOSV) {
-                        if (IS_JDOSV) newCP=932;
-                        else if (IS_PDOSV) newCP=936;
-                        else if (IS_KDOSV) newCP=949;
-                        else if (IS_TDOSV) newCP=950;
+                    if(!newCP && IS_DOSV) {
+                        if(IS_JDOSV) newCP = 932;
+                        else if(IS_PDOSV) newCP = 936;
+                        else if(IS_KDOSV) newCP = 949;
+                        else if(IS_TDOSV) newCP = 950;
+                    }
+                    if((control->opt_langcp && msgcodepage > 0 ) || CheckDBCSCP(msgcodepage)|| msgcodepage == dos.loaded_codepage) newCP = msgcodepage;
+                    if (newCP != dos.loaded_codepage && (!TTF_using() || (TTF_using() && isSupportedCP(newCP)))) {
+                        int missing = toSetCodePage(this, newCP, msgcodepage?-1:control->opt_fastlaunch?1:-2);
+                        //WriteOut(MSG_Get("SHELL_CMD_CHCP_ACTIVE"), dos.loaded_codepage);
+                        if (missing > 0) WriteOut(MSG_Get("SHELL_CMD_CHCP_MISSING"), missing);
+                        else if (missing < 0) WriteOut(MSG_Get("SHELL_CMD_CHCP_INVALID"), newCP);
                     }
-                    const char* name = DOS_GetLoadedLayout();
-                    if (newCP==932||newCP==936||newCP==949||newCP==950||newCP==951) {
-                        dos.loaded_codepage=newCP;
-                        SetupDBCSTable();
-                        runRescan("-A -Q");
-                        DOSBox_SetSysMenu();
-                    } else if (control->opt_langcp && !name && (layout.empty() || layout=="auto"))
-                        SetKEYBCP();
                 }
-                if (lastmsgcp && lastmsgcp != dos.loaded_codepage) SwitchLanguage(lastmsgcp, dos.loaded_codepage, true);
             }
-			if (country>0&&!control->opt_noconfig) {
+            if (country>0&&!control->opt_noconfig) {
 				countryNo = country;
 				DOS_SetCountry(countryNo);
 			}
-			const char * extra = section->data.c_str();
+            runRescan("-A -Q");
+            const char * extra = section->data.c_str();
 			if (extra&&!control->opt_securemode&&!control->SecureMode()&&!control->opt_noconfig) {
 				std::string vstr;
 				std::istringstream in(extra);
@@ -990,10 +977,12 @@ void DOS_Shell::Prepare(void) {
         internal_program = true;
 		VFILE_Register("4DOS.INI",(uint8_t *)i4dos_data,(uint32_t)strlen(i4dos_data), "/4DOS/");
         internal_program = false;
-        unsigned int cp=dos.loaded_codepage;
-        if (!dos.loaded_codepage) InitCodePage();
-        initcodepagefont();
-        dos.loaded_codepage=cp;
+        //unsigned int cp=dos.loaded_codepage;
+        //if (!dos.loaded_codepage) InitCodePage();
+        //initcodepagefont();
+        //dos.loaded_codepage=cp;
+        finish_prepare = true;
+
     }
 #if (defined(WIN32) && !defined(HX_DOS) || defined(LINUX) && C_X11 || defined(MACOSX)) && (defined(C_SDL2) || defined(SDL_DOSBOX_X_SPECIAL))
     if (enableime) SetIMPosition();
@@ -1093,38 +1082,38 @@ public:
         int cp=dos.loaded_codepage;
         InitCodePage();
         force_conversion = false;
-        int ind=0;
 
         /* The user may have given .BAT files to run on the command line */
         if (!control->auto_bat_additional.empty()) {
             std::string cmd = "@echo off\n";
 
-            for (unsigned int i=0;i<control->auto_bat_additional.size();i++) {
+            decltype(control->auto_bat_additional)::size_type ind = 0;
+            for (auto & str : control->auto_bat_additional) {
                 if (!control->opt_prerun) cmd += "\n";
-                if (!strncmp(control->auto_bat_additional[i].c_str(), "@mount c: ", 10)) {
-                    cmd += control->auto_bat_additional[i]+"\n";
+                if (!strncmp(str.c_str(), "@mount c: ", 10)) {
+                    cmd += str+"\n";
                     cmd += "@config -get lastmount>nul\n";
                     cmd += "@if not '%CONFIG%'=='' %CONFIG%";
                 } else {
                     std::string batname;
-                    //LOG_MSG("auto_bat_additional %s\n", control->auto_bat_additional[i].c_str());
+                    //LOG_MSG("auto_bat_additional %s\n", str.c_str());
 
-                    std::replace(control->auto_bat_additional[i].begin(),control->auto_bat_additional[i].end(),'/','\\');
+                    std::replace(str.begin(),str.end(),'/','\\');
                     size_t pos = std::string::npos;
                     bool lead = false;
-                    for (unsigned int j=0; j<control->auto_bat_additional[i].size(); j++) {
+                    for (unsigned int j=0; j<str.size(); j++) {
                         if (lead) lead = false;
-                        else if ((IS_PC98_ARCH && shiftjis_lead_byte(control->auto_bat_additional[i][j])) || (isDBCSCP() && isKanji1(control->auto_bat_additional[i][j]))) lead = true;
-                        else if (control->auto_bat_additional[i][j]=='\\') pos = j;
+                        else if ((IS_PC98_ARCH && shiftjis_lead_byte(str[j])) || (isDBCSCP() && isKanji1(str[j]))) lead = true;
+                        else if (str[j]=='\\') pos = j;
                     }
                     if(pos == std::string::npos) {  //Only a filename, mount current directory
-                        batname = control->auto_bat_additional[i];
+                        batname = str;
                         cmd += "@mount c: . -nl -q\n";
                     } else { //Parse the path of .BAT file
-                        std::string batpath = control->auto_bat_additional[i].substr(0,pos+1);
+                        std::string batpath = str.substr(0,pos+1);
                         if (batpath==".\\") batpath=".";
                         else if (batpath=="..\\") batpath="..";
-                        batname = control->auto_bat_additional[i].substr(pos+1);
+                        batname = str.substr(pos+1);
                         cmd += "@mount c: \"" + batpath + "\" -nl -q\n";
                     }
                     std::string opt = control->opt_o.size() > ind && control->opt_o[ind].size() ? " "+control->opt_o[ind] : "";
@@ -1150,7 +1139,6 @@ public:
                 }
                 if (control->opt_prerun) cmd += "\n";
             }
-
             autoexec_auto_bat.Install(cmd);
         }
         dos.loaded_codepage = cp;
@@ -1275,7 +1263,12 @@ public:
 #else
 		if (secure) autoexec[i++].Install("z:\\system\\config.com -securemode");
 #endif
-
+#if defined(WIN32)
+        if(TTF_using()) {
+            autoexec[i++].Install("@config -set output=surface");
+            autoexec[i++].Install("@config -set output=ttf");
+        }
+#endif
 		if (addexit) autoexec[i++].Install("exit");
 
 		assert(i <= 17); /* FIXME: autoexec[] should not be fixed size */
@@ -1297,7 +1290,7 @@ static void AUTOEXEC_ShutDown(Section * sec) {
 	}
     if (first_shell != NULL) {
 		delete first_shell;
-		first_shell = 0;//Make clear that it shouldn't be used anymore
+		first_shell = nullptr;//Make clear that it shouldn't be used anymore
     }
     if (call_shellstop != 0) {
         CALLBACK_DeAllocate(call_shellstop);
@@ -1765,12 +1758,25 @@ void SHELL_Init() {
 	uint16_t env_seg;//=DOS_FIRST_SHELL+19; //DOS_GetMemory(1+(4096/16))+1;
 	uint16_t stack_seg;//=DOS_GetMemory(2048/16,"COMMAND.COM stack");
     uint16_t tmp,total_sz;
+    bool tiny_memory_mode = false;
+
+    // below a certain memory size, alter memory arrangement and allocation to minimize memory
+    if (MEM_TotalPages() < 0x8) tiny_memory_mode = true;
 
     // decide shell env size
-    if (dosbox_shell_env_size == 0)
-        dosbox_shell_env_size = (0x158u - (0x118u + 19u)) << 4u; /* equivalent to mainline DOSBox */
-    else
+    if (dosbox_shell_env_size == 0) {
+        if (MEM_TotalPages() >= 0x10/*64KB or more*/)
+            dosbox_shell_env_size = (0x158u - (0x118u + 19u)) << 4u; /* equivalent to DOSBox SVN */
+        else if (MEM_TotalPages() >= 0x8/*32KB or more*/)
+            dosbox_shell_env_size = 384;
+        else if (MEM_TotalPages() >= 0x4/*16KB or more*/)
+            dosbox_shell_env_size = 256;
+        else
+            dosbox_shell_env_size = 144;
+    }
+    else {
         dosbox_shell_env_size = (dosbox_shell_env_size+15u)&(~15u); /* round up to paragraph */
+    }
 
     LOG_MSG("COMMAND.COM env size:             %u bytes",dosbox_shell_env_size);
 
@@ -1793,9 +1799,20 @@ void SHELL_Init() {
     LOG_MSG("COMMAND.COM environment block:    0x%04x sz=0x%04x",env_seg,tmp);
 
     // COMMAND.COM main binary (including PSP and stack)
-    tmp = 0x1A + (2048/16);
+    if (tiny_memory_mode)
+        tmp = 0x13 + (1536/16);
+    else
+        tmp = 0x1A + (2048/16);
     total_sz = tmp;
-	if (!DOS_AllocateMemory(&psp_seg,&tmp)) E_Exit("COMMAND.COM failed to allocate main body + PSP segment");
+
+    // Use normal MCB allocation unless memsize is 4KB
+    if (MEM_TotalPages() > 1) {
+        if (!DOS_AllocateMemory(&psp_seg,&tmp)) E_Exit("COMMAND.COM failed to allocate main body + PSP segment");
+    }
+    else {
+        psp_seg = DOS_GetMemory(tmp,"COMMAND.COM main body and PSP");
+    }
+
     LOG_MSG("COMMAND.COM main body (PSP):      0x%04x sz=0x%04x",psp_seg,tmp);
 
     DOS_SetMemAllocStrategy(savedMemAllocStrategy);
@@ -1817,12 +1834,21 @@ void SHELL_Init() {
     }
 
     // set the stack at 0x1A
-    stack_seg = psp_seg + 0x1A;
+    if (tiny_memory_mode)
+        stack_seg = psp_seg + 0x13;
+    else
+        stack_seg = psp_seg + 0x1A;
     LOG_MSG("COMMAND.COM stack:                0x%04x",stack_seg);
 
     // set the stack pointer
 	SegSet16(ss,stack_seg);
-	reg_sp=2046;
+
+    if (tiny_memory_mode)
+        reg_sp=1534;
+    else
+        reg_sp=2046;
+
+    LOG(LOG_MISC,LOG_DEBUG)("Shell init SS:SP %04x:%04x",(unsigned int)stack_seg,(unsigned int)reg_sp);
 
 	/* Set up int 24 and psp (Telarium games) */
 	real_writeb(psp_seg+16+1,0,0xea);		/* far jmp */
@@ -1970,7 +1996,7 @@ void SHELL_Run() {
 	try {
 		first_shell->Run();
 		delete first_shell;
-		first_shell = 0;//Make clear that it shouldn't be used anymore
+		first_shell = nullptr;//Make clear that it shouldn't be used anymore
 		prepared = false;
 		dos_shell_running_program = false;
 #if DOSBOXMENU_TYPE == DOSBOXMENU_HMENU
@@ -1979,7 +2005,7 @@ void SHELL_Run() {
 	}
 	catch (...) {
 		delete first_shell;
-		first_shell = 0;//Make clear that it shouldn't be used anymore
+		first_shell = nullptr;//Make clear that it shouldn't be used anymore
 		prepared = false;
 		dos_shell_running_program = false;
 #if DOSBOXMENU_TYPE == DOSBOXMENU_HMENU
